/*
 * USB RedRat3 IR Transceiver driver - 0.3
 *
 * This driver is based on the lirc_mceusb driver packaged with
 * the lirc 0.7.2 distribution, which in turn was based upon the
 * usb skeleton driver within the kernel, and the notice from
 * that package has been retained below.
 *
 * Lirc mceusb 0.7.2 Driver Copyright (c) 2003-2004 Dan Conti
 * (dconti@acm.wwu.edu)
 *
 * With thanks to Chris Dodge for his assistance in details for
 * this driver.
 *
 * The RedRat3 is a USB tranceiver with both send & receive,
 * with 2 seperate sensors available for receive to enable
 * both good long range reception for general use, and good
 * short range reception when required for learning a signal.
 *
 * http://www.redrat.co.uk/
 *
 * It uses its own little protocol to communicate, the required
 * parts of which are embedded within this driver.
 *
 * Although originally based upon lirc_mceusb, this resultant
 * driver evolved into something quite different. It has been
 * tested as working with the authors system for the purposes
 * of send, and recv, and recording new remotes, and xmode2
 *
*/

/**********************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 5)
#error "*******************************************************"
#error "Sorry, this driver needs kernel version 2.6.5 or higher"
#error "*******************************************************"
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#endif

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <linux/smp_lock.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <asm/uaccess.h>
#else
#include <linux/uaccess.h>
#endif
#include <linux/usb.h>
#include <linux/mutex.h>
#include <linux/completion.h>

#include "drivers/kcompat.h"
#include "drivers/lirc.h"
#include "drivers/lirc_dev/lirc_dev.h"

#define LIRC_VALUE_MASK 0x00FFFFFF
#define LIRC_MODE2_PULSE 0x01000000

/* Driver Information */
#define DRIVER_VERSION "0.50"
#define DRIVER_AUTHOR "The Dweller,Stephen Cox"
#define DRIVER_DESC "USB RedRat3 IR Transceiver Driver"
#define DRIVER_NAME "lirc_redrat3"

/* module parameters */
#ifdef CONFIG_USB_DEBUG
static int debug = 1;
#else
static int debug;
#endif
#define dprintk(fmt, args...)					\
	do {							\
		if (debug)					\
			printk(KERN_DEBUG __FILE__ ": " \
			DRIVER_NAME ": " fmt "\n", ## args);	\
	} while (0)


#define RR3_MAX_BUF_SIZE       4096
/* transfer buffer size on bulk transfers */
#define RR3_TX_BUFFER_LEN     4096

#define RR3_ERROR			   0x01
#define RR3_MOD_SIGNAL_IN      0x20
#define RR3_MOD_SIGNAL_OUT     0x21

/* Get the RR firmware version */
#define RR3_FW_VERSION         0xB1
/* Send encoded signal bulk-sent earlier*/
#define RR3_TX_SEND_SIGNAL     0xB3
#define RR3_SET_IR_PARAM       0xB7
#define RR3_GET_IR_PARAM       0xB8
/* Start capture with the RC receiver */
#define RR3_RC_DET_ENABLE      0xBB
/* Stop capture with the RC receiver */
#define RR3_RC_DET_DISABLE     0xBC
/* Return the status of RC detector capture */
#define RR3_RC_DET_STATUS      0xBD
/* Reset redrat */
#define RR3_RESET              0xA0

/* Max number of lengths in the signal. */
#define RR3_IR_IO_MAX_LENGTHS  0x01
/* Periods to measure mod. freq. */
#define RR3_IR_IO_PERIODS_MF   0x02
/* Size of memory for main signal data */
#define RR3_IR_IO_SIG_MEM_SIZE 0x03
/* Delta value when measuring lengths */
#define RR3_IR_IO_LENGTH_FUZZ  0x04
/* Timeout for end of signal detection */
#define RR3_IR_IO_SIG_TIMEOUT  0x05
/* Minumum value for pause recognition. */
#define RR3_IR_IO_MIN_PAUSE    0x06

/* Clock freq. of EZ-USB chip */
#define RR3_CLK                24000000
/* Clock periods per timer count */
#define RR3_CLK_PER_COUNT      12
/* (RR3_CLK / RR3_CLK_PER_COUNT) */
#define RR3_CLK_CONV_FACTOR    2000000

/* Raw Modulated signal data value offsets */
#define RR3_RAW_MSIG_PAUSE_OFFSET       0
#define RR3_RAW_MSIG_FREQ_COUNT_OFFSET  4
#define RR3_RAW_MSIG_NUM_PERIOD_OFFSET  6
#define RR3_RAW_MSIG_MAX_LENGTHS_OFFSET 8
#define RR3_RAW_MSIG_NUM_LENGTHS_OFFSET 9
#define RR3_RAW_MSIG_MAX_SIGS_OFFSET    10
#define RR3_RAW_MSIG_NUM_SIGS_OFFSET    12
#define RR3_RAW_MSIG_REPEATS_OFFSET     14

/* Size of the fixed-length portion of the signal */
#define RR3_RAW_MSIG_HEADER_LENGTH      15

/* Raw Modulated signal data value lengths */
#define RR3_RAW_MSIG_PAUSE_LENGTH       4
#define RR3_RAW_MSIG_FREQ_COUNT_LENGTH  2
#define RR3_RAW_MSIG_NUM_PERIOD_LENGTH  2
#define RR3_RAW_MSIG_MAX_LENGTHS_LENGTH 1
#define RR3_RAW_MSIG_NUM_LEGNTHS_LENGTH 1
#define RR3_RAW_MSIG_MAX_SIGS_LENGTH    2
#define RR3_RAW_MSIG_NUM_SIGS_LENGTH    2
#define RR3_RAW_MSIG_REPEATS_LENGTH     1

/* The 8051's CPUCS Register address */
#define RR3_CPUCS_REG_ADDR     0x7f92

#define RR3_DRIVER_MAXLENS 128

/* Define these values to match your device */
#define USB_RR3USB_VENDOR_ID	0x112a
#define USB_RR3USB_PRODUCT_ID	0x0001
#define USB_RR3IIUSB_PRODUCT_ID 0x0005

#define RR3_TIME_UNIT 50

#define RR3_WBUF_LEN 256

/* we can have up to this number of device plugged in at once */
#define MAX_DEVICES		16

/* table of devices that work with this driver */
static struct usb_device_id redrat3_dev_table[] = {
	/* Original version of the RedRat3 */
	{USB_DEVICE(USB_RR3USB_VENDOR_ID, USB_RR3USB_PRODUCT_ID)},
	/* Second Version/release of the RedRat3  - RetRat3II*/
	{USB_DEVICE(USB_RR3USB_VENDOR_ID, USB_RR3IIUSB_PRODUCT_ID)},
	{}			/* Terminating entry */
};


/* Structure to hold all of our device specific stuff */
struct redrat3_dev {
	/* save off the usb device pointer */
	struct usb_device *udev;
	/* the interface for this device */
	struct usb_interface *interface;
	/* the USB device number */
	unsigned char devnum;

	/* the buffer to receive data */
	unsigned char *bulk_in_buffer;
	/* the size of the receive buffer */
	int bulk_in_size;
	/* usb dma */
	dma_addr_t dma_in;
	dma_addr_t dma_out;
	
	/* urb used to read ir data */
	struct urb *read_urb;
	/* the address of the bulk in endpoint */
	__u8 bulk_in_endpointAddr;

	/* the buffer to send data */
	unsigned char *bulk_out_buffer;
	/* the size of the send buffer */
	int bulk_out_size;
	/* the urb used to send data */
	struct urb *write_urb;
	/* the address of the bulk out endpoint */
	__u8 bulk_out_endpointAddr;

	/* true if write urb is busy */
	bool write_busy;
	/* wait for the write to finish */
	struct completion write_finished;

	/* for timeouts */
	wait_queue_head_t wait_q;

	/* locks this structure */
	struct mutex lock;

	/* if the device is not disconnected */
	int present;

	/* lirc */
	struct lirc_driver *plugin;

	/* Is the device currently receiving? */
	bool rr3_recv_in_progress;
	/* is the detector enabled*/
	bool rr3_det_enabled;
	/* Is the device currently trasnmitting?*/
	bool writing_data;

	/* store for current packet */
	char rr3buf[RR3_MAX_BUF_SIZE];
	uint16_t txLength;
	uint16_t txType;
	uint16_t bytesRead;
	/* indicate whether we are going to reprocess
	 * the USB callback with a bigger buffer */
	int buftoosmall;
	char *dataPtr;

	struct timeval last_time;

	unsigned int freq;

};

/* TBC - index struct to point to the raw buffer for readability
 * lengthVals and data are not complete*/
/* sequence is important! - will represent a memory block*/
static struct redrat3_packet_pointers {
	/*uint16_t* PacketLength;	/* Buf */
	/*uint16_t* SignalType;	/* Buf + 2 */
	uint32_t* IntraSigPause;/* Buf + 4*/
	uint16_t* FreqCount;		/* Buf + 8*/
	uint16_t* NumPeriod;	/* Buf + 10*/
	uint8_t* MaxLengths;	/* Buf + 12*/
	uint8_t* NumLengths;		/* Buf + 13*/
	uint16_t* MaxSignals;	/* Buf + 14*/
	uint16_t* NumSignals;	/* Buf + 16*/
	uint8_t* Repeats;		/* Buf + 18*/
	/* array of lengths - not COMPLETE*/
	uint16_t* lengthVals;	/* Buf + 19*/
	/* array of data - not COMPLETE */
	char* data;			/* Buf + 19 +(count(lengthVals)*len(uint16_t))*/
};

/* stores the values of a redrat3 packet in host encoding */
static struct redrat3_packet_host {
	uint32_t IntraSigPause;
	uint16_t FreqCount;
	uint16_t NumPeriod;
	uint8_t MaxLengths;
	uint8_t NumLengths;
	uint16_t MaxSignals;
	uint16_t NumSignals;
	uint8_t Repeats;
	/*
	 * note these two are dynamically allocated
	 * and will need to be freed afterwards
	 * */
	uint16_t* lengthVals;
	uint8_t* data;

};

static void PopulatePacketPointers(
		struct redrat3_packet_pointers *packet,
		char *buffer) {

	/* init the first member to the start of the buffer
	 * then populate each member with the previous member + its length
	 */
/*	packet->PacketLength = (uint16_t*) buffer;
	packet->SignalType = packet->PacketLength + sizeof(*(packet->PacketLength)); 
	dprintk("SignalType Size %d", sizeof(*(packet->SignalType)));
	packet->IntraSigPause = ((uint32_t*) (packet->SignalType)) + sizeof(*(packet->SignalType)); */
	packet->IntraSigPause =  buffer + RR3_RAW_MSIG_PAUSE_OFFSET;
	packet->FreqCount = buffer + RR3_RAW_MSIG_FREQ_COUNT_OFFSET; 
	packet->NumPeriod = buffer + RR3_RAW_MSIG_NUM_PERIOD_OFFSET; 
	packet->MaxLengths = buffer + RR3_RAW_MSIG_MAX_LENGTHS_OFFSET; 
	packet->NumLengths = buffer + RR3_RAW_MSIG_NUM_LENGTHS_OFFSET; 
	packet->MaxSignals = buffer + RR3_RAW_MSIG_MAX_SIGS_OFFSET; 
	packet->NumSignals = buffer + RR3_RAW_MSIG_NUM_SIGS_OFFSET; 
	packet->Repeats = buffer + RR3_RAW_MSIG_REPEATS_OFFSET; 
	
}


static void DebugDumpPacketValues(struct redrat3_packet_pointers *packet) {
/*	dprintk( "PacketLength: %d", ntohs(*(packet->PacketLength))); 	
	dprintk( "SignalType: %d", ntohs(*(packet->SignalType))); */
	dprintk( "IntraSigPause: %d", ntohl(*(packet->IntraSigPause)));
	dprintk( "FreqCount: %d", ntohs(*(packet->FreqCount)));
	dprintk( "NumPeriod: %d", *(packet->NumPeriod));
	dprintk( "MaxLengths: %d", *(packet->MaxLengths));
	dprintk( "NumLengths: %d", *(packet->NumLengths));
	dprintk( "MaxSignals: %d", ntohs(*(packet->MaxSignals)));
	dprintk( "Repeats: %d", *(packet->Repeats));
}

static void PopulateHostPacketFromPointers(
		struct redrat3_packet_pointers *raw,
		struct redrat3_packet_host *decoded) {
	
	decoded->IntraSigPause = ntohl(*(raw->IntraSigPause));
	decoded->FreqCount = ntohs(*(raw->FreqCount));
	decoded->NumPeriod = *(raw->NumPeriod);
	decoded->MaxLengths = *(raw->MaxLengths);
	decoded->NumLengths = *(raw->NumLengths);
	decoded->MaxSignals = ntohs(*(raw->MaxSignals));
	decoded->NumSignals = ntohs(*(raw->NumSignals));
	decoded->Repeats = *(raw->Repeats);
}


/* helper function declaration */
static int rr3ir_issue_async(struct redrat3_dev *dev);

static void _rr3TranslateFirmwareError(int errCode, int devnum)
{
	dprintk("Entering %s ", __func__);

	switch (errCode) {
	case 0x00:
		dprintk(" Dev[%d] No Error", devnum);
		break;

	/* Codes 0x20 through 0x2F are reserved for IR Firmware Errors */
	case 0x20:
		dprintk(
			" Dev[%d] Initial signal pulse not long enough "
			"to measure carrier frequency.", devnum);
		break;
	case 0x21:
		dprintk(
			" Dev[%d] Not enough length values allocated for"
			" signal.", devnum);
		break;
	case 0x22:
		dprintk(
			" Dev[%d] Not enough memory allocated for signal "
			"data.", devnum);
		break;
	case 0x23:
		dprintk(
			" Dev[%d] Too many signal repeats.", devnum);
		break;
	case 0x28:
		dprintk(
			" Dev[%d] Insufficient memory available for IR signal "
			"data memory allocation.", devnum);
		break;
	case 0x29:
		dprintk(
			" Dev[%d] Insufficient memory available "
			"for IrDa signal data memory allocation.",
			devnum);
		break;

	/* Codes 0x30 through 0x3F are reserved for USB Firmware Errors */
	case 0x30:
		dprintk
		(DRIVER_NAME
			" Dev[%d] Insufficient memory available for bulk "
			"transfer structure."
			, devnum);
		break;

	/* Other error codes... These are primarily errors that can occur in the
	 * control messages sent to the redrat
	 */
	case 0x40:
		dprintk(
			" Dev[%d] Signal capture has been terminated.",
			devnum);
		break;
	case 0x41:
		dprintk(
			" Dev[%d] Attempt to set/get and unknown signal I/O "
			"algorithm parameter.", devnum);
		break;
	case 0x42:
		dprintk(
			" Dev[%d] Signal capture already started.", devnum);
		break;

	default:
		dprintk(" Dev[%d] Unknown Error", devnum);
		break;
	}
}

static uint32_t valToModFreq(uint16_t modFreqCount, uint16_t numPeriods)
{
	uint32_t modFreq = 0;

	dprintk("Entering %s ", __func__);

	if (modFreqCount != 0) {
		modFreq =
		    (RR3_CLK * numPeriods) / (modFreqCount *
					      RR3_CLK_PER_COUNT);
	} else {
		modFreq = 0;
	}

	return modFreq;
}

static uint32_t lengthToMicrosec(uint32_t length)
{

	/*this code scales down the figures for the same result... */
	uint32_t biglen = length * 1000;
	uint32_t divisor = (RR3_CLK_CONV_FACTOR) / 1000;
	uint32_t result = (uint32_t) (biglen / divisor);

	if (result == 0) {
		/*dont allow zero lengths to go back!! lirc breaks! */
		result = 1;
	}

	return result;
}

static uint32_t microsecToLen(uint32_t microsec)
{

/*length * 1000   length * 1000000              */
/*------------- = ---------------- = micro	*/
/*rr3clk / 1000       rr3clk                    */

/*6 * 2       4 * 3        micro * rr3clk          micro * rr3clk / 1000 */
/*----- = 4   ----- = 6    -------------- = len    ---------------------*/
/*  3           2             1000000                    1000           */


	uint32_t biglen = microsec;
	uint32_t divisor = (RR3_CLK_CONV_FACTOR / 1000);
	uint32_t result = (uint32_t) (biglen * divisor) / 1000;

	dprintk("Entering %s ", __func__);

	if (result == 0)
		result = 1;	/*dont allow zero lengths to go back!! */

	return result;

}

static void send_packet_to_lirc(struct redrat3_dev *ir, lirc_t* lircdata)
{
	if ((*lircdata)) {
		lirc_buffer_write(ir->plugin->rbuf,
				  (unsigned char *) lircdata);
		wake_up(&ir->plugin->rbuf->wait_poll);
		(*lircdata) = 0;
	}
}


static void rr3DecodeSignalData(struct redrat3_dev *dev,
			uint8_t *sigData,
			int sigLength)
{
	uint16_t *lengthVals = NULL;
	uint8_t *dataVals = NULL;
	int i = 0;

	uint32_t pause = 0;
	uint16_t modFreqCount = 0;
	uint16_t noPeriods = 0;
	uint8_t maxLengths = 0;
	uint8_t noLengths = 0;
	uint16_t maxSigSize = 0;
	uint16_t sigSize = 0;
	uint8_t noRepeats = 0;

	uint32_t modFreq = 0;
	uint32_t intraSigPause = 0;

	lirc_t singledata = 0;
	uint32_t singleLength = 0;
	struct timeval now;

	struct redrat3_packet_pointers packet;

	dprintk("Entering %s ", __func__);

	/* Sanity check */
	if (!(sigLength >= RR3_RAW_MSIG_HEADER_LENGTH))
		dprintk(
		" Dev[%d] read returned less than rr3 header len",
		dev->devnum);

	PopulatePacketPointers(&packet, sigData);
	DebugDumpPacketValues(&packet);

	memcpy(&pause, sigData, RR3_RAW_MSIG_PAUSE_LENGTH);
	pause = ntohl(pause);

	memcpy(&modFreqCount, sigData + RR3_RAW_MSIG_FREQ_COUNT_OFFSET,
	       RR3_RAW_MSIG_FREQ_COUNT_LENGTH);
	modFreqCount = ntohs(modFreqCount);

	memcpy(&noPeriods, sigData + RR3_RAW_MSIG_NUM_PERIOD_OFFSET,
	       RR3_RAW_MSIG_NUM_PERIOD_LENGTH);
	noPeriods = ntohs(noPeriods);

	maxLengths = sigData[RR3_RAW_MSIG_MAX_LENGTHS_OFFSET];
	noLengths = sigData[RR3_RAW_MSIG_NUM_LENGTHS_OFFSET];

	memcpy(&maxSigSize, sigData + RR3_RAW_MSIG_MAX_SIGS_OFFSET,
	       RR3_RAW_MSIG_MAX_SIGS_LENGTH);
	maxSigSize = ntohs(maxSigSize);

	memcpy(&sigSize, sigData + RR3_RAW_MSIG_NUM_SIGS_OFFSET,
	       RR3_RAW_MSIG_NUM_SIGS_LENGTH);
	sigSize = ntohs(sigSize);

	noRepeats = sigData[RR3_RAW_MSIG_REPEATS_OFFSET];
	
	/* Calculate values */
	modFreq = valToModFreq(modFreqCount, noPeriods);
	intraSigPause = lengthToMicrosec(pause);

	/* Sanity check(s)... Make sure we actually got a well-formed signal
	 * from the redrat... */
	/*assert(sigLength >= (15 + (noLengths * sizeof(uint16_t)))); */

	/*TODO: disabled this assert, as some received packets failed on it. */
	/*assert(sigLength==(15+ (maxLengths * sizeof(uint16_t)) + sigSize));*/

	/* Here we pull out the 'length' values from the signal */
	lengthVals = (uint16_t *) (sigData + RR3_RAW_MSIG_HEADER_LENGTH);
	
	dataVals = (sigData + RR3_RAW_MSIG_HEADER_LENGTH +
		    (maxLengths * sizeof(uint16_t)));

	dprintk(
		"modfreqcount %d, pause %d, modfreq %d, intrasigpause %d",
	     modFreqCount, pause, modFreq, intraSigPause);
	dprintk(
		"noLengths %d, sizSize %d, noRepeats %d",
	     noLengths, sigSize, noRepeats);
	
	/* now process data into the lirc buffer... */
	/* allocate & blank some temporary structures */
	singledata = 0;

	/*write the space since the last data... */
	do_gettimeofday(&now);
	singledata = now.tv_sec - dev->last_time.tv_sec;
	if (singledata + 1 > LIRC_VALUE_MASK / 1000000)
		singledata = LIRC_VALUE_MASK;
	else {
		singledata *= 1000000;
		singledata += now.tv_usec - dev->last_time.tv_usec;
	}
	dev->last_time.tv_sec = now.tv_sec;
	dev->last_time.tv_usec = now.tv_usec;
	
	send_packet_to_lirc(dev, &singledata);
	
	/* process each rr3 encoded byte into a lirc_t */
	for (i = 0; i < sigSize; i++) {
		/*dprintk("  length index [%d] is %d",i,data[i]); */
		/* set lirc_t length */
		singleLength = lengthToMicrosec((uint32_t) ntohs(lengthVals[dataVals[i]]));

		/*cap the value to LIRC_VALUE_MASK */
		if (singleLength > LIRC_VALUE_MASK)
			singleLength = LIRC_VALUE_MASK;
		
		dprintk(" i:%d\t datavals[i]:%d \t lengths[dataVals[i]]:%d ", 
				i, dataVals[i], singleLength);

		singledata = singleLength;
		if (i % 2) {
			/* spaces dont need any additional work..
			 * we use this opportunity to 'tune' the
			 * detected data a little, to compensate for
			 * the detection hardware
			 */
			singledata += 130;
			if (singledata > LIRC_VALUE_MASK)
				singledata = LIRC_VALUE_MASK;
		} else {
			/* pulses need LIRC_MODE2_PULSE set...
			 * we use this opportunity to 'tune' the
			 * detected data a little, to compensate for
			 * the detection hardware
			 */
			singledata = (singledata | LIRC_MODE2_PULSE);
			singledata -= 120;
		}
		send_packet_to_lirc(dev, &singledata);
	}

	return;
}


/* Util fn to send rr3 cmds */
static uint8_t rr3_send_cmd(int rr3cmd, struct usb_device *udev)
{
	uint8_t retval;
	int res = 0;
	dprintk("Entering %s ", __func__);

	res = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
			      rr3cmd,
			      USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			      USB_DIR_IN, 0x0000, 0x0000, &retval, 1,
			      HZ * 10);

	if (res >= 0) {
		/*dprintk("rr3_send_cmd: \
		 * Success sending rr3 cmd res %d, retval %d",
		 *  res, retval);*/
		return retval;
	} else {
		dprintk("rr3_send_cmd: \
		  Error sending rr3 cmd res %d, retval %d",
		  res, retval);
		return -1;
	}

}

/* Enables the rr3 long range detector,
 * and starts async receive */
static int rr3_enable_detector(struct redrat3_dev *dev)
{
	uint8_t ret = 1;
	dprintk("Entering %s ", __func__);
	if (!dev->rr3_det_enabled) {
		ret = rr3_send_cmd(RR3_RC_DET_STATUS, dev->udev);
		dprintk(
			" Dev[%d] current detector status... %d",
			dev->devnum, ret);

		dprintk(
			" Dev[%d] enabling detector",
			dev->devnum);
		ret = rr3_send_cmd(RR3_RC_DET_ENABLE, dev->udev);
		if (ret != 0) {
			dprintk(" rr3: problem enabling detector");
			/*return 0; */
		} else {
			dprintk
			    (DRIVER_NAME
			    " setting detector as enabled within driver");
			dev->rr3_det_enabled = true;
			dprintk(" issuing async read against rr3");
			rr3ir_issue_async(dev);
		}

		ret = rr3_send_cmd(RR3_RC_DET_STATUS, dev->udev);
		dprintk(" detector status... %d", ret);
	} else {
		dprintk
		(DRIVER_NAME
		" dev[%d] not enabling detector "
		", driver state:enabled already", dev->devnum);
	}
	return 1;
}

/* Disables the rr3 long range detector */
static int rr3_disable_detector(struct redrat3_dev *dev)
{
	uint8_t ret = 1;
	dprintk("Entering %s ", __func__);

	if (dev->rr3_det_enabled) {
		ret = rr3_send_cmd(RR3_RC_DET_STATUS, dev->udev);
		dprintk(" current detector status... %d", ret);

		ret = rr3_send_cmd(RR3_RC_DET_DISABLE, dev->udev);
		if (ret != 0)
			dprintk(" problem disabling detector");


		ret = rr3_send_cmd(RR3_RC_DET_STATUS, dev->udev);
		dprintk(" detector status... %d", ret);

		dev->rr3_det_enabled = (ret != 0);
	} else {
		dprintk
			(DRIVER_NAME
			" not disabling detector, driver: disabled already");
	}
	return 1;
}

static inline void rr3usb_delete(struct redrat3_dev *dev)
{
	dprintk("Entering %s ", __func__);

	/*rr3_disable_detector(dev); */

	usb_buffer_free(dev->udev, dev->bulk_in_size, dev->bulk_in_buffer,
			dev->dma_in);
	usb_buffer_free(dev->udev, dev->bulk_out_size,
			dev->bulk_out_buffer, dev->dma_out);

	if (dev->write_urb != NULL)
		usb_free_urb(dev->write_urb);

	dev->rr3_det_enabled = false;

	/*TODO: properly handle this.. */
	if (dev->read_urb != NULL) {
		/*usb_kill_urb( dev->read_urb ); */
		usb_unlink_urb(dev->read_urb);
		usb_free_urb(dev->read_urb);
	}

	kfree(dev);
}

static void rr3usb_reset(struct usb_device *udev)
{
	int rc = 0;
	uint8_t flag = 0;
	uint32_t value32 = 0;
	uint16_t value16 = 0;
	uint8_t value8 = 0;

	dprintk("Entering %s ", __func__);

	flag = 0x01;
	rc = usb_control_msg(udev,
			     usb_rcvctrlpipe(udev, 0),
			     RR3_RESET,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_IN, RR3_CPUCS_REG_ADDR, 0, &flag,
			     sizeof flag, HZ * 25);
	dprintk(" reset returned 0x%02X ", rc);

	value8 = 5;
	rc = usb_control_msg(udev,
			     usb_sndctrlpipe(udev, 0),
			     RR3_SET_IR_PARAM,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_OUT, RR3_IR_IO_LENGTH_FUZZ, 0,
			     &value8, sizeof(value8), HZ * 25);
	dprintk(" set ir parm len fuzz %d rc 0x%02X", value8, rc);
	value8 = RR3_DRIVER_MAXLENS;
	rc = usb_control_msg(udev,
			     usb_sndctrlpipe(udev, 0),
			     RR3_SET_IR_PARAM,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_OUT, RR3_IR_IO_MAX_LENGTHS, 0,
			     &value8, sizeof(value8), HZ * 25);
	dprintk(
		" set ir parm max lens %d rc 0x%02X",
		value8, rc);


	/*TMP code to dump params in device... */
	value32 = 0;
	rc = usb_control_msg(udev,
			     usb_rcvctrlpipe(udev, 0),
			     RR3_GET_IR_PARAM,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_IN, RR3_IR_IO_SIG_TIMEOUT, 0,
			     &value32, sizeof(value32), HZ * 5);
	dprintk(" io sig timeout returned 0x%02X value %d", rc,
		value32);
	
	value8 = 0;
	rc = usb_control_msg(udev,
			     usb_rcvctrlpipe(udev, 0),
			     RR3_GET_IR_PARAM,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_IN, RR3_IR_IO_LENGTH_FUZZ, 0, &value8,
			     sizeof(value8), HZ * 5);
	dprintk("io len fuzz returned 0x%02X value %d",
		rc, value8);
	
	value8 = 0;
	rc = usb_control_msg(udev,
			     usb_rcvctrlpipe(udev, 0),
			     RR3_GET_IR_PARAM,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_IN, RR3_IR_IO_MAX_LENGTHS, 0, &value8,
			     sizeof(value8), HZ * 5);
	dprintk(" io max len returned 0x%02X value %d", rc, value8);

	value16 = 0;
	rc = usb_control_msg(udev,
			     usb_rcvctrlpipe(udev, 0),
			     RR3_GET_IR_PARAM,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_IN, RR3_IR_IO_SIG_MEM_SIZE, 0,
			     &value16, sizeof(value16), HZ * 5);
	dprintk(" io sigmem size returned 0x%02X value %d", rc,
		value16);
}

static void rr3usb_setup(struct usb_device *udev)
{
	int rc = 0;
	char *buffer;
	int size = 256;
	dprintk("Entering %s ", __func__);

	buffer = kzalloc(sizeof(char) * (size + 1), GFP_KERNEL);

	rr3usb_reset(udev);

	rc = usb_control_msg(udev,
			     usb_rcvctrlpipe(udev, 0),
			     RR3_FW_VERSION,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			     USB_DIR_IN, 0, 0, buffer, size, HZ * 5);

	if (rc >= 0)
		dprintk(" Firmware : %s", buffer);
	else
		dprintk(" Problem with firmware ID");

	kfree(buffer);
	return;
}

static void rrir_cleanup(struct redrat3_dev *dev)
{
	dprintk("Entering %s ", __func__);
	memset(dev->bulk_in_buffer, 0, dev->bulk_in_size);
}

static int redrat3_ir_open(void *data)
{
	struct redrat3_dev *dev = (struct redrat3_dev *) data;
	dprintk("Entering %s ", __func__);

	if (!dev) {
		printk(DRIVER_NAME "[?]: %s called with no context\n",
		       __func__);
		return -EIO;
	}

	dprintk
	    (DRIVER_NAME " Dev[%d] Set use inc, setting driver "
			"state to reading in progress", dev->devnum);
	dev->rr3_recv_in_progress = true;
	dprintk(" Dev[%d] enabling detector", dev->devnum);
	rr3_enable_detector(dev);

	MOD_INC_USE_COUNT;
	return 0;
}

static void redrat3_ir_close(void *data)
{
	struct redrat3_dev *dev = (struct redrat3_dev *) data;
	dprintk("Entering %s ", __func__);

	if (!dev) {
		printk(DRIVER_NAME "[?]: %s called with no context\n",
		       __func__);
		return;
	}

	dprintk
	    (DRIVER_NAME " Dev[%d] Set use dec, reverting driver "
			"state from reading in progress", dev->devnum);

	dev->rr3_recv_in_progress = false;

	MOD_DEC_USE_COUNT;
}


/*
 * rr3ir_issue_async
 *
 *  Issues an async read to the ir data in port..
 *  sets the callback to be rr3ir_handle_async
 */
static int rr3ir_issue_async(struct redrat3_dev *dev)
{
	int res = 0;
	dprintk("Entering %s ", __func__);

	if (dev->rr3_det_enabled) {

		dev->read_urb->dev = dev->udev;	/*reset the urb->dev ptr */

		memset(dev->bulk_in_buffer, 0, dev->bulk_in_size);
		res = usb_submit_urb(dev->read_urb, GFP_ATOMIC);
		if (res) {
			dprintk
				(" issue of async receive request"
				 " FAILED! (res=%d)\n",
			     res);
			dprintk(
				" urb->transfer_buffer_length (%d)\n",
				dev->read_urb->transfer_buffer_length);
			return -1;
		}

		dprintk(
			" after issue of async read (res=%d)\n", res);
	} else {
		dprintk(" not issuing async read, "
			"as detector not enabled");
	}

	return 0;
}

static void rr3_readUSBPacket_Start(int len, struct redrat3_dev *dev)
{
	/* return code */
	int rc;
	uint16_t txError = 0;
	
	dprintk("Entering %s ", __func__);
	dprintk(" processing as start of rr3 packet");

	/* grab the Length and type of transfer */
	memcpy(&(dev->txLength), (unsigned char *) dev->bulk_in_buffer,
	       sizeof(dev->txLength));
	memcpy(&(dev->txType), ((unsigned char *) dev->bulk_in_buffer +
		sizeof(dev->txLength)),
	       sizeof(dev->txType));

	/*data needs conversion to know what its real values are*/
	dev->txLength = ntohs(dev->txLength);
	dev->txType = ntohs(dev->txType);

	if (dev->txType == RR3_ERROR) {
		memcpy(&txError, ((unsigned char *)dev->bulk_in_buffer
			+ (sizeof(dev->txLength) + sizeof(dev->txType))),
		       sizeof(txError));
		txError = ntohs(txError);
		dprintk(" txError data was 0x%02x", txError);
		_rr3TranslateFirmwareError (txError, dev->devnum);
	}
	else if (dev->txType == RR3_MOD_SIGNAL_IN) {
		dev->bytesRead = len;
		dev->bytesRead -=
		    (sizeof(dev->txLength) +
		     sizeof(dev->txType));
		dev->dataPtr = &(dev->rr3buf[0]);

		/*cant process the packet if
		 * we cant read it all, */
		if (dev->txLength >
		    RR3_MAX_BUF_SIZE)
			dev->buftoosmall = 1;

		if (!dev->buftoosmall) {
			memcpy(dev->dataPtr,
			       ((unsigned char *)
				dev->bulk_in_buffer
				+
				sizeof
				dev->txLength +
				sizeof
				dev->txType),
			       dev->bytesRead);
			dev->dataPtr +=
			    dev->bytesRead;
		}
		/*dprintk("bytesRead %d,
		 * txLength %d",dev->bytesRead,
		 * dev->txLength);*/

		if (dev->bytesRead < dev->txLength) {
			dprintk(" not got complete packet yet..."
				" resumbitting async read urb");
			rc = rr3ir_issue_async(dev);
			dprintk(" Resubmit urb returned %d", rc);
			return;
		}
	}
	else {
		dprintk
			(" ignoring this packet with type 0x%02x,"
			" len of %d, 0x%02x",
		     dev->txType, len,
		     dev->txLength);
	} 
}

static void rr3_readUSBPacket_Continuation(int len, struct redrat3_dev *dev)
{
	
	dprintk("Entering %s ", __func__);

	/* only attempt to clone the
	 * data to our local buffer
	 * if it was large enough,
	 * else we just try to skip
	 * the remainder of this packet */
	if (!dev->buftoosmall) {
		memcpy(dev->dataPtr,
		       (unsigned char *)
		       dev->bulk_in_buffer, len);
		dev->dataPtr += len;
	}

	dev->bytesRead += len;
	dprintk(" buffer now at %d/%d",
		dev->bytesRead, dev->txLength);
}

static void ProcessValidUSB(int len, struct redrat3_dev *dev)
{

	dprintk("Entering %s ", __func__);

	/*dprintk("processing valid usb return"); */
	if (dev->bytesRead == 0
	    && len >=
	    (sizeof(dev->txType) +
	     sizeof(dev->txLength))) {
	     	rr3_readUSBPacket_Start(len, dev);
	}
	/*reading continuation bytes */
	else if (dev->bytesRead != 0) {
		rr3_readUSBPacket_Continuation(len, dev);
	} else if (dev->bytesRead == 0) {
		/*implication that len is less than
		 * sizeof txlen txtype */
		/*hmm.. got less bytes than minimum */
		dprintk(" Didn't get"
		" minimum length packet");
	}

	if (dev->bytesRead > dev->txLength) {
		dprintk
			(" reading %d more bytes than expected",
		     (dev->bytesRead - dev->txLength));
		dev->bytesRead = 0;
		dev->txLength = 0;
		dev->txType = 0;
	} else if (dev->bytesRead == dev->txLength) {
		/*if we get here, our read size matches
		 * the expected packet size */

		if (dev->txType == RR3_MOD_SIGNAL_IN) {
			/*this method will decode rr3
			 * to mode2 data and append it
			 * to the dev->lircdata buf till full*/
			if (dev->rr3_recv_in_progress) {
				dprintk
				(" type was "
				" mod_signal_in,"
				" passing off to rr3"
				" decode method");
				rr3DecodeSignalData(dev,
					dev->rr3buf,
					dev->txLength);
			} else {
				dprintk(" discarding rr3 packet as device"
						" not recording");
			}
		} else {
			dprintk
				(" txType was not mod_signal_in,"
				" discarding packet with"
				" type 0x%02X",dev->txType);
		}
		dev->bytesRead = 0;
		dev->txLength = 0;
		dev->txType = 0;
	}
}


/*
 * callback function from USB
 * when async USB request has completed
 *
 */
static void rr3ir_handle_async(struct urb *urb, struct pt_regs *regs)
{
	struct redrat3_dev *dev;
	dprintk("Entering %s ", __func__);

	dprintk("USB async callback");

	if (!urb) {
		dprintk(
		" urb was null - cant clear the read flag!");
		return;
	}

	dev = urb->context;
	if (dev) {

		if (!dev->rr3_det_enabled) {
			dprintk
				(" received a read callback but detector"
				" disabled - ignoring");
			return;
		}

		/* USB_ST_NOERROR==0 */
		if (urb->status == 0) {	
			ProcessValidUSB(urb->actual_length, dev);
		} else {
			dprintk(" urb->status (%d) was bad",
			     urb->status);
			/*len wasnt big enough to contain the rr3 info..
			 * so we ignore this packet
			 * setting recv to 0 means the
			 * add_to_buf call can reschedule reads..
			 * clear of recv moved to fn exit */
			dev->bytesRead = 0;
			dev->txLength = 0;
			dev->txType = 0;
		}

	}

	if (!dev->writing_data)
		rr3ir_issue_async(dev);
	else
		dprintk(" Not reissuing read,"
			" as write in progress");
}

static int rr3ir_alloc_urb(struct redrat3_dev *dev)
{
	dprintk("Entering %s ", __func__);

	dev->read_urb = usb_alloc_urb(0, GFP_KERNEL);
	dprintk(" Allocated dev->read_urb="
		"usb_alloc_urb(0, GFP_KERNEL)");


	if (!dev->read_urb) {
		err("No free urbs for read available");
		return -1;
	}

	usb_fill_bulk_urb(dev->read_urb, dev->udev,
			  usb_rcvbulkpipe(dev->udev, 0x82),
			  dev->bulk_in_buffer, dev->bulk_in_size,
			  (usb_complete_t) rr3ir_handle_async, dev);

	return 0;
}


/**
 *	rr3usb_write_bulk_callback
 */
static void rr3usb_write_bulk_callback(struct urb *urb,
				       struct pt_regs *regs)
{
	struct redrat3_dev *dev = (struct redrat3_dev *) urb->context;
	dprintk("Entering %s ", __func__);

	dprintk(" %s - devnum %d", __func__, dev->devnum);

	if ((urb->status != -ENOENT) && (urb->status != -ECONNRESET)) {
		dprintk(" %s - nonzero write"
			"-back status received: %d",
			__func__, urb->status);
		return;
	}

	return;
}

/*There has to be a better way!*/
void debugDumpIoctl(unsigned int cmd)
{
	dprintk("Entering %s ", __func__);

	switch (cmd) {
	case LIRC_GET_FEATURES:
		dprintk(" LIRC_GET_FEATURES - %lu",
			LIRC_GET_FEATURES);
		break;
	case LIRC_GET_SEND_MODE:
		dprintk(" LIRC_GET_SEND_MODE - %lu",
			LIRC_GET_SEND_MODE);
		break;
	case LIRC_GET_REC_MODE:
		dprintk(" LIRC_GET_REC_MODE - %lu",
			LIRC_GET_REC_MODE);
		break;
	case LIRC_GET_SEND_CARRIER:
		dprintk("LIRC_GET_SEND_CARRIER - %lu",
			LIRC_GET_SEND_CARRIER);
		break;
	case LIRC_GET_REC_CARRIER:
		dprintk(" LIRC_GET_REC_CARRIER - %lu",
			LIRC_GET_REC_CARRIER);
		break;
	case LIRC_GET_SEND_DUTY_CYCLE:
		dprintk(" LIRC_GET_SEND_DUTY_CYCLE - %lu",
			LIRC_GET_SEND_DUTY_CYCLE);
		break;
	case LIRC_GET_REC_DUTY_CYCLE:
		dprintk(" LIRC_GET_REC_DUTY_CYCLE - %lu",
			LIRC_GET_REC_DUTY_CYCLE);
		break;
	case LIRC_GET_REC_RESOLUTION:
		dprintk(" LIRC_GET_REC_RESOLUTION - %lu",
			LIRC_GET_REC_RESOLUTION);
		break;
	case LIRC_GET_LENGTH:
		dprintk(" LIRC_GET_LENGTH - %lu",
			LIRC_GET_LENGTH);
		break;
	case LIRC_SET_SEND_MODE:
		dprintk(" LIRC_SET_SEND_MODE - %lu",
			LIRC_SET_SEND_MODE);
		break;
	case LIRC_SET_REC_MODE:
		dprintk(" LIRC_SET_REC_MODE - %lu",
			LIRC_SET_REC_MODE);
		break;
	case LIRC_SET_SEND_CARRIER:
		dprintk(" LIRC_SET_SEND_CARRIER - %lu",
			LIRC_SET_SEND_CARRIER);
		break;
	case LIRC_SET_REC_CARRIER:
		dprintk(" LIRC_SET_REC_CARRIER - %lu",
			LIRC_SET_REC_CARRIER);
		break;
	case LIRC_SET_SEND_DUTY_CYCLE:
		dprintk(" LIRC_SET_SEND_DUTY_CYCLE - %lu",
			LIRC_SET_SEND_DUTY_CYCLE);
		break;
	case LIRC_SET_REC_DUTY_CYCLE:
		dprintk(" LIRC_SET_REC_DUTY_CYCLE - %lu",
			LIRC_SET_REC_DUTY_CYCLE);
		break;
	case LIRC_SET_TRANSMITTER_MASK:
		dprintk(" LIRC_SET_TRANSMITTER_MASK - %lu",
			LIRC_SET_TRANSMITTER_MASK);
		break;
	case LIRC_SET_REC_DUTY_CYCLE_RANGE:
		dprintk(" LIRC_SET_REC_DUTY_CYCLE_RANGE - %lu",
			LIRC_SET_REC_DUTY_CYCLE_RANGE);
		break;
	case LIRC_SET_REC_CARRIER_RANGE:
		dprintk(" LIRC_SET_REC_CARRIER_RANGE - %lu",
			LIRC_SET_REC_CARRIER_RANGE);
		break;
	default:
		dprintk(" bad lirc ioctl - %u", cmd);
	}
}

static int redrat3_lirc_ioctl(struct inode *node, struct file *filep,
		      unsigned int cmd, unsigned long arg)
{
	int result;
	unsigned long value;
	unsigned int ivalue;
	struct redrat3_dev *ir = NULL;

	dprintk("Entering %s ", __func__);

	dprintk(" servicing ioctl request");


	/* Retrieve lirc_driver data for the device */
	ir = lirc_get_pdata(filep);
	if (!ir)
		return -EFAULT;


	switch (cmd) {
	case LIRC_SET_SEND_CARRIER:
		result = get_user(ivalue, (unsigned int *) arg);
		if (result)
			return result;

		dprintk(" SET_SEND_CARRIER old:"
			" %u new :%u \n", ir->freq,
			ivalue);

		if (ir->freq != ivalue)
			ir->freq = ivalue;

		return 0;
		break;

	case LIRC_GET_SEND_MODE:

		result = put_user(LIRC_MODE2SEND(LIRC_MODE_PULSE),
				  (unsigned long *) arg);

		dprintk(" GET SEND MODE %d", result);

		if (result)
			return result;
		break;

	case LIRC_SET_SEND_MODE:

		result = get_user(value, (unsigned long *) arg);

		dprintk(" SET SEND MODE %d", result);

		if (result)
			return result;
		break;

	case LIRC_GET_LENGTH:
		return -ENOSYS;
		break;

	default:
		debugDumpIoctl(cmd);
		return -ENOIOCTLCMD;
	}
	return 0;

}

static void dumpEncodedSigData(char *origbuffer)
{
	int i = 0;
	uint32_t tmpi;
	uint16_t tmps;
	uint8_t tmp8;
	uint8_t lens;
	uint16_t *lengthVals = NULL;
	uint8_t *dataVals = NULL;
	char *buffer = origbuffer + 4;
	dprintk("Entering %s ", __func__);

	dprintk("Signal Dump");

	memcpy(&tmps, origbuffer, sizeof(tmps));
	tmps = ntohs(tmps);
	dprintk(" Size %d", tmps);

	memcpy(&tmps, origbuffer + 2, sizeof(tmps));
	tmps = ntohs(tmps);
	dprintk(" Type 0x%02x", tmps);

	memcpy(&tmpi, buffer, sizeof(tmpi));
	tmpi = ntohl(tmpi);
	dprintk(" Intrasig %d", tmpi);

	memcpy(&tmps, buffer + 4, sizeof(tmps));
	tmps = ntohs(tmps);
	dprintk(" Modfreq %d", tmps);

	memcpy(&tmp8, buffer + 9, 1);
	/*tmp8=ntohl(tmp8); */
	dprintk(" No of lens %d", tmp8);
	lens = tmp8;

	tmps = 0;
	memcpy(&tmps, buffer + 12, 2);
	tmps = ntohs(tmps);
	dprintk(" Sigsize %d", tmps);

	memcpy(&tmp8, buffer + 14, 1);
	/*tmp8=ntohs(tmp8); */
	dprintk(" Repeats %d", tmp8);

	lengthVals = (uint16_t *) (buffer + 15);

	i = lens;
	while (i > 0) {
		dprintk(" len %d", ntohs(lengthVals[lens - i]));
		i--;
	}

	dprintk(" Indexes");

	/*read 128 from ctrl msg */
	dataVals = (buffer + 15 + (128 * sizeof(uint16_t)));

	if (tmps > 256) {
		dprintk(" SIGsize is fubar >256");
		tmps = 256;
	}
	i = tmps;
	while (i > 0) {
		if (i > 10) {
			dprintk
			    (DRIVER_NAME " (%d-%d) %02X %02X %02X %02X %02X"
			     " %02X %02X %02X %02X %02X",
			     tmps - i, tmps - i + 9, dataVals[tmps - i],
			     dataVals[tmps - i + 1],
			     dataVals[tmps - i + 2],
			     dataVals[tmps - i + 3],
			     dataVals[tmps - i + 4],
			     dataVals[tmps - i + 5],
			     dataVals[tmps - i + 6],
			     dataVals[tmps - i + 7],
			     dataVals[tmps - i + 8], dataVals[tmps - i + 9]
			    );
			i -= 10;
		} else if (i > 5) {
			dprintk(" (%d-%d) %02X %02X %02X %02X %02X",
				tmps - i, tmps - i + 5,
				dataVals[tmps - i], dataVals[tmps - i + 1],
				dataVals[tmps - i + 2],
				dataVals[tmps - i + 3],
				dataVals[tmps - i + 4]
			    );
			i -= 5;
		} else {
			dprintk(" (%d) %02X",
				tmps - i, dataVals[tmps - i]);
			i -= 1;
		}
	}

	dprintk(" \n");

}

static uint16_t modFreqToVal(unsigned int modFreq)
{
	int mult = 6000000;
	dprintk("Entering %s ", __func__);

	dprintk(" converting modFreq %du", modFreq);

	/* Clk used in mod. freq. generation is CLK24/4. */

	dprintk(" using mult of %d", mult);
	dprintk(" using subval of %d", (mult / modFreq));

	return (uint16_t) (65536 - (mult / modFreq));
}

static ssize_t redrat3_transmit_ir(struct file *file, const char *buf,
			  size_t n, loff_t *ppos)
{
	struct redrat3_dev *dev = NULL;
	int sendret;
	int retval, i, j, count;
	lirc_t wbuf[RR3_WBUF_LEN];
	lirc_t lircLens[126];	/*rr3 max lengths.. */
	uint8_t *sigdata = NULL;
	int lencheck = 0;
	int curlencheck = 0;
	uint16_t modfreq;
	uint16_t intrasig;

	uint32_t tmpi = 0;
	uint16_t tmps = 0;
	uint8_t tmp8 = 0;
	int resultLength = 0;
	uint16_t *lengthsPtr = NULL;
	uint8_t *dataPtr = NULL;
	/* read this from ctrl msg... */
	uint8_t maxLengths = RR3_DRIVER_MAXLENS;
	int sendBufLength;

	char *buffer;
	
	struct redrat3_packet_pointers packet;
	dprintk("Entering %s ", __func__);

	/* Retrieve lirc_driver data for the device */
	dev = lirc_get_pdata(file);
	if (!dev)
		return -EFAULT;

	if (dev == NULL)
		return -EINVAL;

	if (dev->writing_data) {
		/* write in progress */
		dprintk(" write called when already writing? eh??");
		return -EINVAL;
	}

	if (n % sizeof(lirc_t))
		return -EINVAL;
	retval = !access_ok(VERIFY_READ, buf, n);
	if (retval)
		return retval;


	count = n / sizeof(lirc_t);
	if (count > RR3_WBUF_LEN || count % 2 == 0)
		return -EINVAL;
	copy_from_user(wbuf, buf, n);

	/*down( &dev->sem ); */
	dprintk(" setting writing data to 1");
	dev->writing_data = true;

	dprintk(" requesting detector shutdown");
	rr3_disable_detector(dev);

	if (dev->rr3_det_enabled) {
		dprintk
		    (DRIVER_NAME " hmm.. det still enabled after "
		     "requesting disable, clearing write flag");
		dev->writing_data = false;
		dprintk
		    (DRIVER_NAME " could be in a right pickle now "
		     "with no read waiting anymore...");
		return -EINVAL;
	}

	memset(lircLens, 0, (sizeof(lirc_t) * 126));

	for (i = 0; i < count; i++) {
		/*dprintk("lirc send byte %lu",wbuf[i]); */

		for (lencheck = 0; lencheck < curlencheck; lencheck++) {
			if (lircLens[lencheck] ==
			    microsecToLen(wbuf[i] & LIRC_VALUE_MASK)) {
				/*dprintk("Found match at position %d for"
				 * " index %d value %d",lencheck,i,wbuf[i]); */
				break;
			}
		}
		/*dprintk("after lencheck %d, curlencheck %d, "
		 * "val %d",lencheck,curlencheck,wbuf[i]); */
		if (lencheck == curlencheck) {
			dprintk
			    (DRIVER_NAME " adding length from index %d (%u) "
			     "to 2list@ %d, enc %u",
			     i, wbuf[i], curlencheck,
			     microsecToLen(wbuf[i] & LIRC_VALUE_MASK));
			if (curlencheck < 255) {
				/*now convert the value to a proper
				 * rr3 value.. */
				lircLens[curlencheck] =
				    microsecToLen(wbuf[i] & LIRC_VALUE_MASK);
				curlencheck++;
			} else
				dprintk
					(DRIVER_NAME " Max rr3 lengths \
					exceeded for "
				    "send command.. ");
			/*TODO: add a rescan with increasing
			 *  'fudge factors' to make similar amounts the same. */
		}
	}

	sigdata =
	    kzalloc((sizeof(uint8_t) *
				(count + 1 + 1)),
				GFP_KERNEL);
	sigdata[count] = 127;
	sigdata[count + 1] = 127;
	for (i = 0; i < count; i++) {
		for (j = 0; j < curlencheck; j++) {
			if (lircLens[j] ==
			    microsecToLen(wbuf[i] & LIRC_VALUE_MASK)) {
				sigdata[i] = j;
			}
		}
	}


	for (i = 0; i < curlencheck; i++) {
		dprintk(" Length val %d - encoded %u \
			de-encoded %u", i,
			lircLens[i], lengthToMicrosec(lircLens[i]));
	}
	sendret = 0;

	modfreq = modFreqToVal(dev->freq);
	intrasig = lengthToMicrosec(100);

	sendBufLength =
	    RR3_RAW_MSIG_HEADER_LENGTH + (sizeof(uint16_t) * maxLengths) +
	    (sizeof(uint8_t) * (count + 1 + 1)) + (sizeof(uint16_t) * 2);

	buffer = kzalloc(sendBufLength, GFP_KERNEL);

	dprintk(" Allocated buffer of %d \
		bytes for send", sendBufLength);

	dprintk(" Encoding intrasig of %d", intrasig);
	tmpi = htonl(intrasig);
	memcpy(buffer + 4, &tmpi, sizeof(tmpi));

	dprintk(" Encoding modfreq of %d \
		based from freq %d", modfreq,
		dev->freq);
	tmps = htons(modfreq);
	memcpy(buffer + 4 + 4, &tmps, 2);

	dprintk(" Encoding no of lens of %d", curlencheck);
	tmp8 = curlencheck;
	buffer[9 + 4] = tmp8;

	dprintk(" Encoding sigsize of %d", count + 1 + 1);
	tmps = htons(count + 1 + 1);
	memcpy(buffer + 4 + 12, &tmps, 2);

	/*lirc handles repeats externally.. not rr3 */
	dprintk(" Encoding zero repeats");
	buffer[14 + 4] = 0;


	lengthsPtr = (uint16_t *) (buffer + 4 + 15);
	for (i = 0; i < curlencheck; ++i)
		lengthsPtr[i] = htons(lircLens[i]);

	dataPtr =
	    (uint8_t *) (buffer + 4 + 15 +
			 (sizeof(uint16_t) * maxLengths));
	memcpy(dataPtr, sigdata, (count + 1 + 1));

	/*now put the first two parts of the header into the packet */
	tmps = htons(sendBufLength - (sizeof(uint16_t) * 2));
	memcpy(buffer, &tmps, 2);

	tmps = htons(RR3_MOD_SIGNAL_OUT);	/*rr3 mod sig out */
	memcpy(buffer + 2, &tmps, 2);

	PopulatePacketPointers(&packet, buffer);
	/*dumpEncodedSigData(buffer); */

	tmps =
	    usb_bulk_msg(dev->udev,
			 usb_sndbulkpipe(dev->udev,
					 dev->bulk_out_endpointAddr),
			 buffer, sendBufLength, &resultLength, 10 * HZ);
	dprintk(" Just sent %d bytes, \
		rc from last send %d", resultLength,
		tmps);

	/*now tell rr3 to transmit what we sent it... */
	sendret = 0;
	tmps = 0;
	sendret = usb_control_msg(dev->udev,
					usb_rcvctrlpipe(dev->udev, 0),
					RR3_TX_SEND_SIGNAL,
				  USB_TYPE_VENDOR | USB_RECIP_DEVICE |
				  USB_DIR_IN, 0, 0, &tmps, sizeof(tmps),
				  HZ * 10);

	dprintk(" Told rr3 to send \
		data, rc %d tmps %d", sendret, tmps);
	if (sendret < 0) {
		dprintk
		    (DRIVER_NAME " Error: sendret<0 (Told rr3 to \
		    send data, rc %d tmps %d)",
		     sendret, tmps);
/*        printUSBErrs();*/
	}

	dprintk(" freeing alloc'd buffers");
	kfree(buffer);
	kfree(sigdata);

	dprintk(" clearing writing data flag");
	dev->writing_data = false;

	dprintk(" re-enabling detector");
	rr3_enable_detector(dev);

	/*up( &dev->sem ); */

	return n;
}

static struct file_operations lirc_fops = {
	.owner	= THIS_MODULE,
	.write	= redrat3_transmit_ir,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
	.ioctl	= redrat3_lirc_ioctl,
#else
	.unlocked_ioctl	= redrat3_lirc_ioctl,
#endif
};


/**
 *	redrat3_dev_probe
 *
 *	Called by the usb core when a new device is connected that it
 *	thinks this driver might be interested in.
 */
static int __devinit redrat3_dev_probe(struct usb_interface *interface,
			const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(interface);
	struct usb_host_interface *iface_desc;
	struct redrat3_dev *dev = NULL;
	struct usb_endpoint_descriptor *endpoint;

	struct lirc_driver *plugin;
	struct lirc_buffer *rbuf;

	size_t buffer_size;
	int i;
	int retval = -ENOMEM;
	int minor = 0;
	dprintk("Entering %s ", __func__);

	/* See if the device offered us matches what we can accept */
	if ((udev->descriptor.idVendor != USB_RR3USB_VENDOR_ID) ||
	    ((udev->descriptor.idProduct != USB_RR3USB_PRODUCT_ID) &&
	     (udev->descriptor.idProduct != USB_RR3IIUSB_PRODUCT_ID))) {
		dprintk(" Wrong Vendor/Product IDs");
		return -ENODEV;
	}

	dprintk(" [%d]: Assigned Device Number\n", udev->devnum);
	/* allocate memory for our device state and initialize it */
	dev = kzalloc(sizeof(struct redrat3_dev), GFP_KERNEL);
	if (dev == NULL) {
		err("Out of memory");
		retval = -ENOMEM;
		goto error;
	}

	mutex_init(&dev->lock);
	dev->udev = udev;
	dev->interface = interface;
	dev->devnum = udev->devnum;

	/* default.. will get overridden by any sends with a freq defined */
	dev->freq = 38000;

	/* set up the endpoint information */
	/* check out the endpoints */
	/* use only the first bulk-in and bulk-out endpoints */

	iface_desc = interface->cur_altsetting;

	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;
	if ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) &&
	    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     == USB_ENDPOINT_XFER_BULK)) {
			dprintk(" Dev[%d]we found a "
				"bulk in endpoint with #0x%02X",
				dev->devnum, endpoint->bEndpointAddress);

			/* data comes in on 0x82, 0x81 is for other data... */
			if (endpoint->bEndpointAddress == 0x82) {

				buffer_size = endpoint->wMaxPacketSize;
				dev->bulk_in_size = buffer_size;
				dev->bulk_in_endpointAddr =
				    endpoint->bEndpointAddress;
				dev->bulk_in_buffer = usb_buffer_alloc
				    (udev, buffer_size, GFP_ATOMIC,
				     &dev->dma_in);
				if (!dev->bulk_in_buffer) {
					err("Couldn't allocate bulk_in_buffer");
					goto error;
				}

			}
		}

		if (((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
		     == 0x00)
		    &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     == USB_ENDPOINT_XFER_BULK)) {
			/*dprintk("we found a bulk out endpoint"); */
			dev->write_urb = usb_alloc_urb(0, GFP_KERNEL);
			if (!dev->write_urb) {
				err("No free urbs for write available");
				goto error;
			}
			buffer_size = endpoint->wMaxPacketSize;
			dev->bulk_out_size = buffer_size;
			dev->bulk_out_endpointAddr =
			    endpoint->bEndpointAddress;
			dev->bulk_out_buffer =
			    usb_buffer_alloc(udev, buffer_size, GFP_ATOMIC,
					     &dev->dma_out);
			if (!dev->bulk_out_buffer) {
				err("Couldn't allocate bulk_out_buffer");
				goto error;
			}
			usb_fill_bulk_urb(dev->write_urb, udev,
					  usb_sndbulkpipe
					  (udev,
					   endpoint->bEndpointAddress),
					  dev->bulk_out_buffer,
					  buffer_size,
					  (usb_complete_t)
					  rr3usb_write_bulk_callback,
					  dev);
		}
	}

	if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
		err("Couldn't find both bulk-in and bulk-out endpoints");
		goto error;
	}

	/* init the waitq */
	init_waitqueue_head(&dev->wait_q);

	/* Set up our lirc driver */
	plugin = kzalloc(sizeof(struct lirc_driver), GFP_KERNEL);
	if (!(plugin)) {
		err("out of memory");
		goto error;
	}

	rbuf = kzalloc(sizeof(struct lirc_buffer), GFP_KERNEL);
	if (!(rbuf)) {
		err("out of memory");
		kfree(plugin);
		goto error;
	}

	/* rbuf is memset inside this method... */
	if (lirc_buffer_init(rbuf, sizeof(lirc_t), 4096)) {
		err("out of memory");
		kfree(plugin);
		kfree(rbuf);
		goto error;
	}

	strcpy(plugin->name, DRIVER_NAME " ");
	plugin->minor = -1;
	plugin->code_length = sizeof(lirc_t);
	plugin->features =
	    LIRC_CAN_REC_MODE2 | LIRC_CAN_SEND_PULSE;
	plugin->data = dev;
	plugin->rbuf = rbuf;

	plugin->fops = &lirc_fops;
	plugin->set_use_inc = &redrat3_ir_open;
	plugin->set_use_dec = &redrat3_ir_close;
	/*we no longer need a polling callback */
	plugin->sample_rate = 0;
	plugin->owner = THIS_MODULE;
    minor = lirc_register_driver(plugin);
	if (minor < 0) {
		kfree(plugin);
		lirc_buffer_free(rbuf);
		kfree(rbuf);
		goto error;
	}
	dev->plugin = plugin;
    plugin->minor = minor;

	rrir_cleanup(dev);
	rr3usb_setup(udev);

	dev->bytesRead = 0;
	dev->dataPtr = NULL;
	dev->txType = 0;
	dev->txLength = 0;
	dev->buftoosmall = 0;
	do_gettimeofday(&dev->last_time);

	if (rr3ir_alloc_urb(dev) == -1)
		return -ENOMEM;

	dev->rr3_recv_in_progress = false;
	dev->writing_data = false;
	dev->rr3_det_enabled = false;


	/* we can register the device now, as it is ready */
	usb_set_intfdata(interface, dev);

	return 0;

error:
	rr3usb_delete(dev);
	dev = NULL;
	dprintk(" %s: retval = %x", __func__, retval);

	return retval;
}

/**
 *	redrat3_dev_disconnect
 *
 *	Called by the usb core when the device is removed from the system.
 *
 *	This routine guarantees that the driver will not submit any more urbs
 *	by clearing dev->udev.  It is also supposed to terminate any currently
 *	active urbs.
 */
static void __devexit redrat3_dev_disconnect(struct usb_interface *interface)
{
	struct redrat3_dev *dev;
	struct lirc_driver *d;
	dprintk("Entering %s ", __func__);

	/*int minor;*/
	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	mutex_lock(&dev->lock);

	rr3usb_delete(dev);

	/* unhook lirc things */
	d = dev->plugin;
	lirc_unregister_driver(d->minor);
	lirc_buffer_free(dev->plugin->rbuf);
	kfree(dev->plugin->rbuf);
	kfree(dev->plugin);
	/* terminate an ongoing write */
	if (dev->write_busy) {
		usb_unlink_urb(dev->write_urb);
		wait_for_completion(&dev->write_finished);
	}

	/* prevent device read, write and ioctl */
	dev->present = 0;

	dprintk(" Dev[%d]: RedRat3 IR Transceiver"
		" now disconnected\n", dev->devnum);

	mutex_unlock(&dev->lock);
}

/* USB Suspend handler */
static int redrat3_dev_suspend(struct usb_interface *intf, pm_message_t message)
{
	struct redrat3_dev *ir = usb_get_intfdata(intf);
	dprintk("Entering %s ", __func__);

	printk(DRIVER_NAME "[%d]: suspend\n", ir->devnum);
	usb_kill_urb(ir->read_urb);
	return 0;
}

/* USB Resume Handler */
static int redrat3_dev_resume(struct usb_interface *intf)
{
	struct redrat3_dev *ir = usb_get_intfdata(intf);
	dprintk("Entering %s ", __func__);

	printk(DRIVER_NAME " Dev[%d]: resume\n", ir->devnum);
	if (usb_submit_urb(ir->read_urb, GFP_ATOMIC))
		return -EIO;
	return 0;
}


/* usb specific object needed to register this driver with
 * the usb subsystem */
static struct usb_driver rr3usb_driver = {
    LIRC_THIS_MODULE(.owner = THIS_MODULE)
	.name = DRIVER_NAME,
	.probe = redrat3_dev_probe,
	.disconnect = redrat3_dev_disconnect,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	.suspend =	redrat3_dev_suspend,
	.resume =	redrat3_dev_resume,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 23)
	.reset_resume =	redrat3_dev_resume,
#endif
#endif
	.id_table = redrat3_dev_table
};


/**
 *	Called on kernel module load
 *  Register the redrat3 lirc driver with the
 *  USB subsystem
 */
static int __init redrat3_dev_init(void)
{

	int result;

	dprintk("Entering %s ", __func__);
	dprintk(" RedRat3 Usb Transceiver : " DRIVER_VERSION);
	dprintk(" debug mode enabled\n");
	dprintk("Compile Date: %s Time: %s", __DATE__, __TIME__);

	/* register this driver with the USB subsystem */
	result = usb_register(&rr3usb_driver);
	if (result) {
		err("usb_register failed for the " DRIVER_NAME
		    " driver. error number %d", result);
		return result;
	}
	printk(KERN_INFO DRIVER_NAME ": " DRIVER_DESC " " DRIVER_VERSION "\n");
	printk(KERN_INFO DRIVER_NAME ": " DRIVER_AUTHOR "\n");

	return 0;
}


/**
 *	Called on kernel module unload
 *  De-registers the redrat3 lirc driver with the
 *  USB subsystem
 */
static void __exit redrat3_dev_exit(void)
{
	dprintk("Entering %s ", __func__);
	printk(KERN_INFO DRIVER_NAME ": " DRIVER_DESC
			" " DRIVER_VERSION " Module Exit\n");
	/* deregister this driver with the USB subsystem */
	usb_deregister(&rr3usb_driver);
}

module_init(redrat3_dev_init);
module_exit(redrat3_dev_exit);

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(usb, redrat3_dev_table);

module_param(debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(debug, "Debug enabled or not");

EXPORT_NO_SYMBOLS;
